#define com_maroontress_libcgi_util_ArraySet_IMPORT
#define com_maroontress_libcgi_util_String_IMPORT

#include <string.h>
#include <stdlib.h>
#include <com/maroontress/libcgi/util/ArraySet.h>
#include <com/maroontress/libcgi/util/String.h>

#define TABLE_SIZE (4095u)

struct String {
    unsigned int hash;
    size_t length;
    unsigned char data[1];
};

struct SharedString {
    size_t count;
    struct String string;
};

struct Property {
    unsigned int hash;
    size_t length;
};

static struct ArraySet *table[TABLE_SIZE];

__attribute__((constructor)) void
String_staticInitialize(void)
{
    unsigned int k;

    for (k = 0; k < TABLE_SIZE; ++k) {
	table[k] = NULL;
    }
}

static void
getProperty(const unsigned char *s, /*@out@*/ struct Property *p)
{
    unsigned int h;
    const unsigned char *m = s;

    for (h = 0; *m != 0; ++m) {
	h = (h << 5) - h + *m;
    }
    p->hash = h;
    p->length = m - s;
}

static struct SharedString *
findSharedString(const unsigned char *m, const struct Property *p)
{
    unsigned int i = p->hash % TABLE_SIZE;
    struct ArraySet *set = table[i];
    size_t k, size;

    if (set == NULL)
	return NULL;
    size = ArraySet_getSize(set);
    for (k = 0; k < size; ++k) {
	struct SharedString *ss;
	struct String *s;

	ss = ArraySet_get(set, k);
	s = &ss->string;
	if (s->hash == p->hash
	    && s->length == p->length
	    && strcmp(s->data, m) == 0) {
	    return ss;
	}
    }
    return NULL;
}

static int
addTable(struct SharedString *ss)
{
    unsigned int i = ss->string.hash % TABLE_SIZE;

    if (table[i] == NULL && (table[i] = ArraySet_create()) == NULL)
	return -1;
    return ArraySet_add(table[i], ss);
}

static struct SharedString *
createSharedString(const unsigned char *m, const struct Property *p)
{
    struct SharedString *ss;
    size_t size;

    size = sizeof(*ss) + p->length;
    if ((ss = (struct SharedString *)malloc(size)) == NULL)
	return NULL;
    ss->count = 1;
    ss->string.hash = p->hash;
    ss->string.length = p->length;
    memcpy(ss->string.data, m, p->length + 1);

    if (addTable(ss) < 0) {
	free(ss);
	return NULL;
    }
    return ss;
}

static struct SharedString *
intern(const unsigned char *m, const struct Property *p)
{
    struct SharedString *ss;

    if ((ss = findSharedString(m, p)) != NULL) {
	++(ss->count);
	return ss;
    }
    return createSharedString(m, p);
}

struct String *
String_intern(const unsigned char *m)
{
    struct Property p;
    struct SharedString *ss;

    getProperty(m, &p);
    ss = intern(m, &p);
    return (ss == NULL) ? NULL : &ss->string;
}

static void
freeSharedString(struct SharedString *ss)
{
    unsigned int i = ss->string.hash % TABLE_SIZE;

    if (--(ss->count) > 0)
	return;
    if (ArraySet_remove(table[i], ss) < 0)
	abort();
    free(ss);
}

void
String_free(struct String *s)
{
    struct Property p;
    struct SharedString *ss;

    if (s == NULL)
	return;
    p.hash = s->hash;
    p.length = s->length;
    if ((ss = findSharedString(s->data, &p)) == NULL)
	abort();
    freeSharedString(ss);
}

unsigned char *
String_getData(struct String *s)
{
    return s->data;
}

size_t
String_getLength(struct String *s)
{
    return s->length;
}
